# This file is part of GElib, a C++/CUDA library for group
# equivariant tensor operations. 
# 
# Copyright (c) 2024, Imre Risi Kondor
#
# This Source Code Form is subject to the terms of the Mozilla
# Public License v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.


import torch
from . import gelib_base as gb
from gelib import *


class SO3irrep:
    """
    An element of the group SO(3)
    """

    def __init__(self, l):
        assert(isinstance(l,int))
        self.obj=gb.SO3irrep(l)
        

    # ---- Static constructors -----------------------------------------------------------------------------



    # ---- Operations --------------------------------------------------------------------------------------


    def matrix(self,R):
        assert(isinstance(R,SO3element))
        return self.obj.matrix(gb.SO3element.view(R))


    # ---- I/O ----------------------------------------------------------------------------------------------


    def __repr__(self):
        return self.obj.__repr__()

    



# WE SHOULD REMOVE THE import *  !!!
import torch

#from gelib_base import SO3element 
#from gelib_base import SO3type 
#from gelib_base import SO3bitype 

from gelib.gelib_common import *
#from gelib.gather_map import *
# Removed redundant import of gather_map

from gelib.SO3element import *
from gelib.SO3irrep import *
from gelib.SO3part import *
from gelib.SO3vec import *

from gelib.SO3type import *
# from gelib.SO3weights import *
# from gelib.SO3weightsArr import *
# from gelib.SO3weights import *
# from gelib.SO3mvec import *

from gelib.SO3partArr import *
from gelib.SO3vecArr import *

#from gelib.SO3weightsArr import *
#from gelib.Wigner import *
#from gelib.CGtensor import *

#from gelib.SO3partB import *
#from gelib.SO3vecB import *

#from gelib.SO3partC import *
#from gelib.SO3vecC import *
#from gelib.SO3partArrC import *
#from gelib.SO3vecArrC import *

#from gelib.SO3bipart import *
#from gelib.SO3bipartArr import *
#from gelib.SO3bivec import *
#from gelib.SO3bivecArr import *
# from gelib.SO3weightsArr import *
# from gelib.Wigner import *
# from gelib.CGtensor import *


def CGproduct(x, y, maxl=-1):
    return x.CGproduct(y, maxl)

def DiagCGproduct(x, y, maxl=-1):
    return x.DiagCGproduct(y, maxl)

def CGtransform(x,l=-1):
    return x.CGtransform(l)

def gather(x,gmap,dim=0):
    return x.gather(gmap,dim)




# This file is part of GElib, a C++/CUDA library for group
# equivariant tensor operations. 
# 
# Copyright (c) 2024, Imre Risi Kondor
#
# This Source Code Form is subject to the terms of the Mozilla
# Public License v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import torch
from . import gelib_base as gb


class SO3type:

    def __init__(self,x):
        if isinstance(x,dict):
            self.obj=gb.SO3type(x)


    # ---- I/O ----------------------------------------------------------------------------------------------


    def __repr__(self):
        return self.obj.__repr__()

    def __str__(self):
        return self.obj.__str__()



    



# This file is part of GElib, a C++/CUDA library for group equivariant 
# tensor operations. 
#  
# Copyright (c) 2023, Imre Risi Kondor
#
# This source code file is subject to the terms of the noncommercial 
# license distributed with GElib in the file NONCOMMERICAL.TXT. Commercial 
# use is prohibited. All redistributed versions of this file (in orginal
# or modified form) must retain this copyright notice and must be 
# accompanied by a verbatim copy of the license. 

import torch
from . import gelib_base as gb
from gelib import *


# ----------------------------------------------------------------------------------------------------------
# ---- SO3part ---------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------


class SO3part(torch.Tensor):
    """
    A collection of vectors that transform according to a specific irreducible representation of SO(3).
    The vectors are stacked into a third order tensor. The first index is the batch index, the second
    is m=-l,...,l, and the third index is the fragment index. 
    """

    def __init__(self, _T):
        self=_T
        #super().__init__()


    # ---- Static constructors -----------------------------------------------------------------------------


    @classmethod
    def zeros(self,b,l,n,device='cpu'):
        """
        Create an SO(3)-part consisting of b lots of n vectors transforming according to the l'th irrep of SO(3).
        The vectors are initialized to zero, resulting in an b*(2+l+1)*n dimensional complex tensor of zeros.
        """
        return SO3part(torch.zeros([b,2*l+1,n],dtype=torch.complex64,device=device))

    @classmethod
    def randn(self,b,l,n,device='cpu'):
        """
        Create an SO(3)-part consisting of b lots of n vectors transforming according to the l'th irrep of SO(3).
        The vectors are initialized as random gaussian vectors, resulting in an b*(2+l+1)*n dimensional random
        complex tensor.
        """
        return SO3part(torch.randn([b,2*l+1,n],dtype=torch.complex64,device=device))

    @classmethod
    def spharm(self,l,X,device='cpu'):
        """
        Return the spherical harmonics of the vectors in the tensor (x,y,z)
        """
        assert(X.dim()==3)
        R = SO3part.zeros(X.size(0),l,X.size(2),device=device)
        R.backend().add_spharm(X)
        return R

    @classmethod
    def Fzeros(self,b,l,device='cpu'):
        """
        Create an SO(3)-part corresponding to the l'th matrix in the Fourier transform of a function on SO(3).
        This gives a b*(2+l+1)*(2l+1) dimensional complex tensor. 
        """
        return SO3part(torch.zeros([b,2*l+1,2*l+1],dtype=torch.complex64,device=device))

    @classmethod
    def Frandn(self,b,l,device='cpu'):
        """
        Create an SO(3)-part corresponding to the l'th matrix in the Fourier transform of a function on SO(3).
        This gives a b*(2+l+1)*(2l+1) dimensional complex random tensor. 
        """
        return SO3part(torch.randn([b,2*l+1,2*l+1],dtype=torch.complex64,device=device))

    def zeros_like(self,*args):
        if not args:
            return SO3part(torch.zeros_like(self))
        if len(args)==2:
            assert isinstance(args[0],int)
            assert isinstance(args[1],int)
            dims=list(self.size())
            dims[-2]=2*args[0]+1
            dims[-1]=args[1]
            return SO3part(torch.zeros(dims,dtype=torch.complex64,device=self.device))
        if len(args)==3:
            assert isinstance(args[0],int)
            assert isinstance(args[1],int)
            dims=list(self.size())
            dims[0]=args[0]
            dims[-2]=2*args[1]+1
            dims[-1]=args[2]
            return SO3part(torch.zeros(dims,dtype=torch.complex64,device=self.device))
    
#     @classmethod
#     def randn_like(self):
#         """
#         Create an SO(3)-part consisting of b lots of n vectors transforming according to the l'th irrep of SO(3).
#         The vectors are initialized as random gaussian vectors, resulting in an b*(2+l+1)*n dimensional random
#         complex tensor.
#         """
#         return SO3part(torch.randn(self.size(),dtype=torch.complex64,device=self.device))

    @classmethod
    def randn_like(self,x):
        return SO3part(torch.randn_like(x))

    def backend(self):
        return gb.SO3part.view(self)

                   
    # ---- Access ------------------------------------------------------------------------------------------


    def getb(self):
        return self.size(0)

    def getl(self):
        return int((self.size(1)-1)/2)

    def getn(self):
        return self.size(2)


    # ---- Operations --------------------------------------------------------------------------------------


    def apply(self, R):
        assert(isinstance(R,SO3element))
        rho=SO3irrep(self.getl())
        return SO3part(torch.matmul(rho.matrix(R),self))


    # ---- Products -----------------------------------------------------------------------------------------


    def odot(self,y):
            return torch.Tensor(torch.sum(torch.mul(torch.view_as_real(self),torch.view_as_real(y))))

    def CGproduct(self, y, l):
        """
        Compute the l component of the Clesbsch--Gordan product of this SO3part with another SO3part y.
        """
        assert isinstance(y,SO3part)
        assert isinstance(l,int)
        return SO3part_CGproductFn.apply(self,y,l)

    def DiagCGproduct(self, y, l):
        """
        Compute the l component of the diagonal Clesbsch--Gordan product of this SO3part with another SO3part y.
        """
        assert isinstance(y,SO3part)
        assert isinstance(l,int)
        return SO3part_DiagCGproductFn.apply(self,y,l)


    # ---- I/O ----------------------------------------------------------------------------------------------

    def __repr__(self):
        return self.backend().__repr__()

    def __str__(self):
        return self.backend().__str__()


# ----------------------------------------------------------------------------------------------------------
# ---- Autograd functions -----------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------


class SO3part_CGproductFn(torch.autograd.Function):

    @staticmethod
    def forward(ctx,x,y,l):
        ctx.l=l
        ctx.save_for_backward(x,y)
        #b=max(x.size(0),y.size(0))
        #r=x.zeros_like(l,x.size(-1)*y.size(-1))
        b=common_batch(x,y)
        r=SO3part.zeros(b,l,x.size(-1)*y.size(-1),device=x.device)
        r.backend().add_CGproduct(x.backend(),y.backend())
        return r

    @staticmethod
    def backward(ctx,g):
        x,y = ctx.saved_tensors
        #g=SO3part(_g)
        xg=x.zeros_like()
        yg=y.zeros_like()
        gb.SO3part.view(xg).add_CGproduct_back0(gb.SO3part.view(g),gb.SO3part.view(y))
        gb.SO3part.view(yg).add_CGproduct_back1(gb.SO3part.view(g),gb.SO3part.view(x))
        return xg,yg,None


class SO3part_DiagCGproductFn(torch.autograd.Function):

    @staticmethod
    def forward(ctx,x,y,l):
        ctx.l=l
        ctx.save_for_backward(x,y)
        #b=max(x.size(0),y.size(0))
        #r=x.zeros_like(l,x.size(-1))
        b=common_batch(x,y)
        assert x.size(-1)==y.size(-1)
        r=SO3part.zeros(b,l,x.size(-1),device=x.device)
        r.backend().add_DiagCGproduct(x.backend(),y.backend())
        return r

    @staticmethod
    def backward(ctx,g):
        x,y = ctx.saved_tensors
        xg=x.zeros_like()
        yg=y.zeros_like()
        gb.SO3part.view(xg).add_DiagCGproduct_back0(gb.SO3part.view(g),gb.SO3part.view(y))
        gb.SO3part.view(yg).add_DiagCGproduct_back1(gb.SO3part.view(g),gb.SO3part.view(x))
        return xg,yg,None


# ----------------------------------------------------------------------------------------------------------
# ---- Other functions --------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------


#def CGproduct(x, y, maxl=-1):
#    return x.CGproduct(y, maxl)


#def DiagCGproduct(x, y, maxl=-1):
#    return x.DiagCGproduct(y, maxl)





# This file is part of GElib, a C++/CUDA library for group equivariant 
# tensor operations. 
#  
# Copyright (c) 2024, Imre Risi Kondor
#
# This source code file is subject to the terms of the noncommercial 
# license distributed with GElib in the file NONCOMMERICAL.TXT. Commercial 
# use is prohibited. All redistributed versions of this file (in orginal
# or modified form) must retain this copyright notice and must be 
# accompanied by a verbatim copy of the license. 

import torch
from . import gelib_base as gb
from gelib import *


# ----------------------------------------------------------------------------------------------------------
# ---- SO3vec ----------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------


class SO3vec:
    """
    An SO(3)-covariant vector consisting of a sequence of SO3part objects, each transforming according
    to a specific irrep of SO(3).
    """

    def __init__(self,*args):
        self.parts={}
        if not args:
            return
        for x in args:
            assert isinstance(x,torch.Tensor)
            p=SO3part(x)
            self.parts[p.getl()]=x
            

    # ---- Static constructors ------------------------------------------------------------------------------


    @classmethod
    def zeros(self,b,tau,device='cpu'):
        "Construct a zero SO3vec object of given type _tau."
        R=SO3vec()
        if isinstance(tau,dict):
            for l,n in tau.items():
                R.parts[l]=SO3part.zeros(b,l,n,device=device)
        return R

    @classmethod
    def randn(self,b,tau,device='cpu'):
        """
        Construct a random SO3vec object of given type _tau.
        >>> import gelib
        >>> import torch
        >>> g = torch.manual_seed(0)
        >>> v = gelib.SO3vec.randn(1,[2,3,1])
        >>> v
        <GElib::SO3vecB of type (2,3,1) [b=1]>
        >>> print(v)
        Part l=0:
          [ (1.08965,-0.207486) (-1.54064,0.401942) ]
        <BLANKLINE>
        <BLANKLINE>
        Part l=1:
          [ (-0.251821,0.309163) (1.02082,0.188125) (0.117702,0.618281) ]
          [ (-0.101451,-0.0789197) (-0.433869,0.0223394) (1.41774,0.0379977) ]
          [ (0.437032,-0.291895) (-0.594723,-1.63769) (-0.0723438,0.560342) ]
        <BLANKLINE>
        <BLANKLINE>
        Part l=2:
          [ (-0.151452,-0.305448) (-0.500545,-0.0752602) (-0.878744,-0.336747) (-0.485018,-1.0643) (0.1803,1.2758) ]
        <BLANKLINE>
        <BLANKLINE>
        """
        R=SO3vec()
        if isinstance(tau,dict):
            for l,n in tau.items():
                R.parts[l]=SO3part.randn(b,l,n,device=device)
        return R

    @classmethod
    def Fzeros(self,b,tau,device='cpu'):
        "Construct a zero SO3vec object of given type _tau."
        R=SO3vec()
        if isinstance(tau,dict):
            for l,n in tau.items():
                R.parts[l]=SO3part.Fzeros(b,l,n,device=device)
        return R

    @classmethod
    def Frandn(self,b,tau,device='cpu'):
        "Construct a zero SO3vec object of given type _tau."
        R=SO3vec()
        if isinstance(tau,dict):
            for l,n in tau.items():
                R.parts[l]=SO3part.Frandn(b,l,n,device=device)
        return R

    @classmethod
    def spharm(self,l,X,device='cpu'):
        """
        Return the spherical harmonics of the vectors in the tensor X
        """
        assert(X.dim()==3)
        R=SO3vec()
        R.parts[l]=SO3part.spharm(b,l,X,device=device)
        return R

    @classmethod
    def zeros_like(self, x):
        R=SO3vec()
        for l,p in x.parts.items():
            R.parts[l]=SO3part.zeros_like(p)
        return R

    @classmethod
    def randn_like(self, x):
        R=SO3vec()
        for l,p in x.parts.items():
            R.parts[l]=SO3part.randn_like(p)
        return R

    def backend(self):
        return gb.SO3vec.view(self.parts)


    # ---- Access -------------------------------------------------------------------------------------------


    def getb(self):
        return self.parts[min(self.parts)].getb()

    def tau(self):
        "Return the 'type' of the SO3vec, i.e., how many components it has corresponding to l=0,1,2,..."
        r={}
        for l,p in self.parts.items():
            r[l]=p.getn()
        return r

    def get_type(self):
        "Return the 'type' of the SO3vec, i.e., how many components it has corresponding to l=0,1,2,..."
        r={}
        for l,p in self.parts.items():
            r[l]=p.getn()
        return r

    def requires_grad_(self):
        for l,p in self.parts.items():
            p.requires_grad_()

    def get_grad(self):
        return SO3vec(*[p.grad for l,p in self.parts.items()])


    # ---- Operations ---------------------------------------------------------------------------------------


    def apply(self, R):
        "Apply the group element to this vector"
        r = SO3vec()
        for l,p in self.parts.items():
            r.parts[l]=p.apply(R)
        return r

    def odot(self,y):
        assert(list(self.parts.keys())==list(y.parts.keys()))
        return sum([self.parts[l].odot(y.parts[l]) for l in self.parts.keys()])

    def __add__(self,y):
        assert(list(self.parts.keys())==list(y.parts.keys()))
        return SO3vec(*[self.parts[l]+y.parts[l] for l in self.parts.keys()])

        
    # ---- Products -----------------------------------------------------------------------------------------


    def CGproduct(self, y, maxl=-1):
        """
        Compute the full Clesbsch--Gordan product of this SO3vec with another SO3vec y.
        """
        xparts=list(self.parts.values())
        yparts=list(y.parts.values())
        rparts =list(SO3vec_CGproductFn.apply(len(xparts), len(yparts), maxl,*(xparts+yparts)))
        return SO3vec(*rparts)

    def DiagCGproduct(self, y, maxl=-1):
        """
        Compute the diagonal Clesbsch--Gordan product of this SO3vec with another SO3vec y.
        """
        xparts=list(self.parts.values())
        yparts=list(y.parts.values())
        rparts =list(SO3vec_DiagCGproductFn.apply(len(xparts), len(yparts), maxl,*(xparts+yparts)))
        return SO3vec(*rparts)


    # ---- I/O ----------------------------------------------------------------------------------------------

    def __repr__(self):
        return self.backend().__repr__()

    def __str__(self):
        return self.backend().__str__()


# ----------------------------------------------------------------------------------------------------------
# ---- Autograd functions -----------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------


class SO3vec_CGproductFn(torch.autograd.Function):

    def __init__(self):
        self.is_sparse=False

    @staticmethod
    def forward(ctx, k1, k2, maxl, *args):
        ctx.k1 = k1
        ctx.k2 = k2
        ctx.save_for_backward(*args)

        x=gb.SO3vec.view(args[0:k1])
        y=gb.SO3vec.view(args[k1:k1+k2])
        b=common_batch(args[0],args[k1])
        tau=x.get_tau().CGproduct(y.get_tau(),maxl)
        rparts=MakeZeroSO3parts(b,tau.get_parts(),args[0].device) # just use regular constructor?
        r=gb.SO3vec.view(rparts)
        r.addCGproduct(x,y)

        return tuple(rparts)


    @staticmethod
    def backward(ctx, *args):

        k1 = ctx.k1
        k2 = ctx.k2
        grads=[torch.zeros_like(x) for x in ctx.saved_tensors]

        x=gb.SO3vec.view(ctx.saved_tensors[0:k1])
        y=gb.SO3vec.view(ctx.saved_tensors[k1:k1+k2])
        g=gb.SO3vec.view(args)
        xg=gb.SO3vec.view(grads[0:k1])
        yg=gb.SO3vec.view(grads[k1:k1+k2])
        xg.addCGproduct_back0(g,y)
        yg.addCGproduct_back1(g,x)

        return tuple([None,None,None]+grads)


class SO3vec_DiagCGproductFn(torch.autograd.Function):

    def __init__(self):
        self.is_sparse=False

    @staticmethod
    def forward(ctx, k1, k2, maxl, *args):
        ctx.k1 = k1
        ctx.k2 = k2
        ctx.save_for_backward(*args)

        x=gb.SO3vec.view(args[0:k1])
        y=gb.SO3vec.view(args[k1:k1+k2])
        b=common_batch(args[0],args[k1])
        tau=x.get_tau().DiagCGproduct(y.get_tau(),maxl)
        rparts=MakeZeroSO3parts(b,tau.get_parts(),args[0].device)
        r=gb.SO3vec.view(rparts)
        r.addDiagCGproduct(x,y)

        return tuple(rparts)

    @staticmethod
    def backward(ctx, *args):

        k1 = ctx.k1
        k2 = ctx.k2
        grads=[torch.zeros_like(x) for x in ctx.saved_tensors]

        x=gb.SO3vec.view(ctx.saved_tensors[0:k1])
        y=gb.SO3vec.view(ctx.saved_tensors[k1:k1+k2])
        g=gb.SO3vec.view(args)
        xg=gb.SO3vec.view(grads[0:k1])
        yg=gb.SO3vec.view(grads[k1:k1+k2])
        xg.addDiagCGproduct_back0(g,y)
        yg.addDiagCGproduct_back1(g,x)

        return tuple([None,None,None]+grads)


class SO3vec_FproductFn(torch.autograd.Function):

    @staticmethod
    def forward(ctx, k1, k2, _maxl, *args):
        ctx.k1 = k1
        ctx.k2 = k2
        ctx.save_for_backward(*args)

        b = args[0].size(0)
        if _maxl == -1:
            maxl = k1+k2-2
        else:
            maxl = _maxl

        r = makeZeroSO3Fparts(b, maxl, args[0].device)

        _x = _SO3vecB.view(args[0:k1])
        _y = _SO3vecB.view(args[k1:k1+k2])
        _r = _SO3vecB.view(r)
        _r.addFproduct(_x, _y)

        return tuple(r)

    @staticmethod
    def backward(ctx, *args):

        k1 = ctx.k1
        k2 = ctx.k2
        # maxl=ctx.maxl

        inputs = ctx.saved_tensors
        assert len(inputs) == k1+k2, "Wrong number of saved tensors."

        grads = [None, None, None]
        for i in range(k1+k2):
            grads.append(torch.zeros_like(inputs[i]))

        _x = _SO3vecB.view(inputs[0:k1])
        _y = _SO3vecB.view(inputs[k1:k1+k2])

        _g = _SO3vecB.view(args)
        _xg = _SO3vecB.view(grads[3:k1+3])
        _yg = _SO3vecB.view(grads[k1+3:k1+k2+3])

        _xg.addFproduct_back0(_g, _y)
        _yg.addFproduct_back1(_g, _x)

        return tuple(grads)


class SO3vec_FmodsqFn(torch.autograd.Function):

    @staticmethod
    def forward(ctx, k1, _maxl, *args):
        ctx.k1 = k1
        # ctx.k2=k1
        ctx.save_for_backward(*args)

        b = args[0].size(0)
        if _maxl == -1:
            maxl = k1+k1-2
        else:
            maxl = _maxl

        r = makeZeroSO3Fparts(b, maxl, args[0].device)

        _x = _SO3vecB.view(args[0:k1])
        _r = _SO3vecB.view(r)
        _r.addFproduct(_x, _x)

        return tuple(r)

    @staticmethod
    def backward(ctx, *args):

        k1 = ctx.k1
        # k2=ctx.k2

        inputs = ctx.saved_tensors
        assert len(inputs) == k1, "Wrong number of saved tensors."

        grads = [None, None]
        for i in range(k1):
            grads.append(torch.zeros_like(inputs[i]))

        _x = _SO3vecB.view(inputs[0:k1])

        _g = _SO3vecB.view(args)
        _xg = _SO3vecB.view(grads[2:k1+2])

        _xg.addFproduct_back0(_g, _x)
        _xg.addFproduct_back1(_g, _x)

        return tuple(grads)


class SO3vec_iFFTFn(torch.autograd.Function):

    @staticmethod
    def forward(ctx, N, *args):

        _v=_SO3vecB.view(args)
        b=_v.getb()
        #maxl=_v.get_maxl()
        ctx.save_for_backward(*args)
        
        r=torch.zeros([b,2*N,N,2*N,2],device=args[0].device)
        _r=ctensorb.view(r)
        _v.add_iFFT_to(_r)

        return r

    @staticmethod
    def backward(ctx, fg):

        inputs = ctx.saved_tensors
        grads = [None]
        for inp in inputs:
            grads.append(torch.zeros_like(inp))

        _fg = ctensorb.view(fg)
        _vg=_SO3vecB.view(grads[1:])
        _vg.add_FFT(_fg)
        
        return tuple(grads)


class SO3vec_FFTFn(torch.autograd.Function):

    @staticmethod
    def forward(ctx, maxl, f):

        ctx.save_for_backward(f)
        _f = ctensorb.view(f)

        v = makeZeroSO3Fparts(_f.get_dim(0), maxl, _f.get_dev())
        _v=_SO3vecB.view(v)
        _v.add_FFT(_f)

        return tuple(v)

    @staticmethod
    def backward(ctx, vg):

        inputs = ctx.saved_tensors

        fg=torch.zeros_like(inputs[0])
        _fg=ctensorb.view(fg)
        _vg.add_iFFT_to(_fg)
        
        return tuple([None, fg])



# ----------------------------------------------------------------------------------------------------------
# ---- Other functions --------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------


def CGproduct(x, y, maxl=-1):
    return x.CGproduct(y, maxl)


def DiagCGproduct(x, y, maxl=-1):
    return x.DiagCGproduct(y, maxl)


def Fproduct(x, y, maxl=-1):
    return x.Fproduct(y, maxl)


def Fmodsq(x, a=-1):
    return x.Fmodsq(a)


# ---- Helpers -----------------------------------------------------------------------------------------------


# def tau_type(x):
#     r = []
#     for t in x:
#         r.append(t.size(2))
#     return r


# def CGproductType(x, y, maxl=-1):
#     if maxl == -1:
#         maxl = len(x)+len(y)-2
#     maxl = min(maxl, len(x)+len(y)-2)
#     r = [0]*(maxl+1)
#     for l1 in range(0, len(x)):
#         for l2 in range(0, len(y)):
#             for l in range(abs(l1-l2), min(l1+l2, maxl)+1):
#                 r[l] += x[l1]*y[l2]
#     return r


# def DiagCGproductType(x, y, maxl=-1):
#     if maxl == -1:
#         maxl = len(x)+len(y)-2
#     maxl = min(maxl, len(x)+len(y)-2)
#     r = [0]*(maxl+1)
#     for l1 in range(0, len(x)):
#         for l2 in range(0, len(y)):
#             for l in range(abs(l1-l2), min(l1+l2, maxl)+1):
#                 r[l] += x[l1]
#     return r


# def DDiagCGproductType(x, maxl=-1):
#     if maxl == -1:
#         maxl = len(x)+(1-len(x)%2)-1
#     maxl = min(maxl, len(x)+(1-len(x)%2)-1)
#     r = [0]*(maxl+1)
#     for l in range(0, len(x)):
#         r[l+l%2] += x[l]
#     return r


def MakeZeroSO3parts(b, tau, device):
    R=[]
    for l,n in tau.items():
        R.append(SO3part.zeros(b,l,n,device))
        #R.append(torch.zeros([b,2*l+1,tau[l]],dtype=torch.cfloat,device=device))
    return R


def makeZeroSO3Fparts(b, maxl, device):
    R = []
    for l,n in tau.items():
        R.append(SO3part.Fzeros(b,l,device))
    return R


def SO3FFT(f,maxl):
    r=SO3vec()
    r.parts=list(SO3vec_FFTFn.apply(maxl,f))
    return r


def SO3iFFT(v,N):
    return v.iFFT(N)





# This file is part of GElib, a C++/CUDA library for group
# equivariant tensor operations. 
# 
# Copyright (c) 2022, Imre Risi Kondor
#
# This Source Code Form is subject to the terms of the Mozilla
# Public License v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.


import torch
from . import gelib_base as gb
import gelib
from gelib import *


class SO3partArr(torch.Tensor):
    """
    A collection of vectors that transform according to a specific irreducible representation of SO(3).
    The vectors are stacked into a fourth order tensor. The first index is the site index, the second index
    is the batch index, the third is m=-l,...,l, and the fourth index is the fragment index. 
    """
    def __init__(self, _T):
        self=_T


    ## ---- Static constructors -----------------------------------------------------------------------------

    
    @staticmethod
    def zeros(b,adims,l,n,device='cpu'):
        """
        Create an SO(3)-part consisting of N*b lots of n vectors transforming according to the l'th irrep of SO(3).
        The vectors are initialized to zero, resulting in an b*(2+l+1)*n dimensional complex tensor of zeros.
        """        
        return SO3partArr(torch.zeros([b]+adims+[2*l+1,n],dtype=torch.complex64,device=device))

    
    @staticmethod
    def ones(b,adims,l,n,device='cpu'):
        """
        Create an SO(3)-part consisting of N*b lots of n vectors transforming according to the l'th irrep of SO(3).
        The vectors are initialized to one, resulting in an b*(2+l+1)*n dimensional complex tensor of ones.
        """        
        return SO3partArr(torch.ones([b]+adims+[2*l+1,n],dtype=torch.complex64,device=device))


    @staticmethod
    def randn(b,adims,l,n,device='cpu'):
        """
        Create an SO(3)-part consisting of N*b lots of n vectors transforming according to the l'th irrep of SO(3).
        The vectors are initialized as random gaussian vectors, resulting in an b*(2+l+1)*n dimensional random
        complex tensor.
        """
        return SO3partArr(torch.randn([b]+adims+[2*l+1,n],dtype=torch.complex64,device=device))


    @classmethod
    def spharm(self, l, X, device='cpu'):
        """
        Return the spherical harmonics of the vector (x,y,z)
        """
        assert(X.size(-2)==3)
        R=SO3partArr.zeros(X.size(0),list(X.size())[1:X.dim()-2],l,X.size(-1),device='cpu')
        R.backend().add_spharm(X)
        return R.to(device)


    @staticmethod
    def Fzeros(b,adims,l,device='cpu'):
        """
        Create an SO(3)-part corresponding to the l'th matrix in the Fourier transform of a function on SO(3).
        This gives a N*b*(2+l+1)*(2l+1) dimensional complex tensor. 
        """
        return SO3partArr(torch.zeros([b]+adims+[2*l+1,2*l+1],dtype=torch.complex64,device=device))


    @staticmethod
    def Frandn(b,adims,l,device='cpu'):
        """
        Create an SO(3)-part corresponding to the l'th matrix in the Fourier transform of a function on SO(3).
        This gives a b*(2+l+1)*(2l+1) dimensional complex random tensor. 
        """
        return SO3partArr(torch.randn([b]+adims+[2*l+1,2*l+1],dtype=torch.complex64,device=device))


    def zeros_like(self,*args):
        if not args:
            return SO3partArr(torch.zeros_like(self))
        if len(args)==2:
            assert isinstance(args[0],int)
            assert isinstance(args[1],int)
            dims=list(self.size())
            dims[-2]=2*args[0]+1
            dims[-1]=args[1]
            return SO3partArr(torch.zeros(dims,dtype=torch.complex64,device=self.device))
    
    
    @classmethod
    def randn_like(self,x):
        return SO3partArr(torch.randn_like(x))


    def backend(self):
        return gb.SO3part.view(self)

    def collapse(self):
        return SO3part(self.reshape(self.size(0),-1,self.size(-2),self.size(-1)))


    ## ---- Access ------------------------------------------------------------------------------------------


    def getb(self):
        return self.size(0)

    def get_adims(self):
        return list(self.size()[1:self.dim()-2])

    def get_nadims(self):
        return self.dim()-3

    def getl(self):
        return int((self.size(-2)-1)/2)

    def getn(self):
        return self.size(-1)


    ## ---- Operations --------------------------------------------------------------------------------------


    def odot(self,y):
        return torch.Tensor(torch.sum(torch.mul(torch.view_as_real(self),torch.view_as_real(y))))

    def apply(self, R):
        assert(isinstance(R,SO3element))
        rho=SO3irrep(self.getl())
        return SO3partArr(torch.matmul(rho.matrix(R),self.collapse()).reshape(self.size()))

    def gather(self,gmap,dim=0):
        """
        Gather the cells of this SO3partArr into a new SO3partArr according to the gather_map
        """
        return SO3partArr_GatherFn.apply(self,gmap,dim)

#     def conterpolate(self,M):
#         return SO3partArr_ConterpolateFn.apply(self,M)

#     def conterpolateB(self,M):
#         return SO3partArr_ConterpolateBFn.apply(self,M)


    # ---- Products -----------------------------------------------------------------------------------------


    def CGproduct(self,y,l):
        """
        Compute the l component of the Clesbsch--Gordan product of this SO3partArr with another SO3partArr y.
        """
        return gelib.SO3partArr_CGproductFn.apply(self,y,l)


    def DiagCGproduct(self,y,l):
        """
        Compute the l component of the diagonal Clesbsch--Gordan product of this SO3partArr with another SO3partArr y.
        """
        return gelib.SO3part_DiagCGproductFn.apply(self,y,l)

    ## ---- I/O ----------------------------------------------------------------------------------------------

        
    def __repr__(self):
        return self.backend().__repr__()

    def __str__(self):
        return self.backend().__str__()


## ----------------------------------------------------------------------------------------------------------
## ---- Autograd functions -----------------------------------------------------------------------------------
## ----------------------------------------------------------------------------------------------------------


class SO3partArr_CGproductFn(torch.autograd.Function):

    @staticmethod
    def forward(ctx,x,y,l):
        ctx.l=l
        ctx.save_for_backward(x,y)
        #b=max(x.size(0),y.size(0))
        #r=x.zeros_like(l,x.size(-1)*y.size(-1))
        b=common_batch(x,y)
        a=x.get_adims()
        r=SO3partArr.zeros(b,a,l,x.size(-1)*y.size(-1),device=x.device)
        r.backend().add_CGproduct(x.backend(),y.backend())
        return r

    @staticmethod
    def backward(ctx,g):
        x,y = ctx.saved_tensors
        #g=SO3part(_g)
        xg=x.zeros_like()
        yg=y.zeros_like()
        gb.SO3part.view(xg).add_CGproduct_back0(gb.SO3part.view(g),gb.SO3part.view(y))
        gb.SO3part.view(yg).add_CGproduct_back1(gb.SO3part.view(g),gb.SO3part.view(x))
        return xg,yg,None


class SO3partArr_DiagCGproductFn(torch.autograd.Function):

    @staticmethod
    def forward(ctx,x,y,l):
        ctx.l=l
        ctx.save_for_backward(x,y)
        #b=max(x.size(0),y.size(0))
        #r=x.zeros_like(l,x.size(-1))
        b=common_batch(x,y)
        a=x.get_adims()
        assert x.size(-1)==y.size(-1)
        r=SO3partArr.zeros(b,a,l,x.size(-1),device=x.device)
        r.backend().add_DiagCGproduct(x.backend(),y.backend())
        return r

    @staticmethod
    def backward(ctx,g):
        x,y = ctx.saved_tensors
        xg=x.zeros_like()
        yg=y.zeros_like()
        gb.SO3part.view(xg).add_DiagCGproduct_back0(gb.SO3part.view(g),gb.SO3part.view(y))
        gb.SO3part.view(yg).add_DiagCGproduct_back1(gb.SO3part.view(g),gb.SO3part.view(x))
        return xg,yg,None


class SO3partArr_GatherFn(torch.autograd.Function): 

    @staticmethod
    def forward(ctx,x,gmap,d):
        ctx.gmap=gmap
        ctx.d=d
        ctx.adims=x.get_adims()
        new_adims=ctx.adims
        new_adims[d]=gmap.d_out()
        r=SO3partArr.zeros(x.getb(),ctx.new_adims,x.getl(),x.getn(),device=x.device)
        r.backend().add_gather(x,gmap,d)
        return r

    @staticmethod
    def backward(ctx,rg):
        xg=SO3partArr.zeros(rg.getb(),ctx.adims,rg.getl(),rg.getn(),device=rg.device)
        xg.backend().add_gather_back(rg,ctx.gmap,ctx.d)
        return xg,None,None


class SO3partArr_ConterpolateFn(torch.autograd.Function): 

    @staticmethod
    def forward(ctx,x,M):
        ctx.nadims=x.get_nadims()
        ctx.M=M
        _x=_SO3partB_array.view(x)
        _M=_rtensor.view(M)
        if(x.get_nadims()==2):
            r=SO3partArr.zeros(x.getb(),x.get_adims()+list(M.size()[:-2]),x.getl(),x.getn(),x.device)
            _r=_SO3partB_array.view(r)
            _r.add_conterpolate2d(_x,_M)
        if(x.get_nadims()==3):
            r=SO3partArr.zeros(x.getb(),x.get_adims()+list(M.size()[:-3]),x.getl(),x.getn(),x.device)
            _r=_SO3partB_array.view(r)
            _r.add_conterpolate3d(_x,_M)
        return r

    @staticmethod
    def backward(ctx,g):
        _g=_SO3partB_array.view(g)
        _M=_rtensor.view(ctx.M)
        gx=SO3partArr.zeros(g.getb(),g.get_adims()[:ctx.nadims],g.getl(),g.getn(),g.device)
        _gx=_SO3partB_array.view(gx)
        if(ctx.nadims==2):
            _gx.add_conterpolate2d_back(_g,_M)
        if(ctx.nadims==3):
            _gx.add_conterpolate3d_back(_g,_M)
        return gx,None


class SO3partArr_ConterpolateBFn(torch.autograd.Function): 

    @staticmethod
    def forward(ctx,x,M):
        ctx.b=x.getb()
        ctx.adims=x.get_adims()
        ctx.l=x.getl()
        ctx.n=x.getn()
        ctx.M=M
        _x=_SO3partB_array.view(x)
        _M=_rtensor.view(M)
        r=torch.zeros([x.getb()]+x.get_adims()+list(M.size()[:-4])+[x.getn()],dtype=torch.cfloat,device=x.device)
        _r=_ctensor.view(r)
        gelib_base.add_conterpolate3dB(_r,_x,_M)
        return r

    @staticmethod
    def backward(ctx,g):
        _g=_ctensor.view(g)
        _M=_rtensor.view(ctx.M)
        gx=SO3partArr.zeros(ctx.b,ctx.adims,ctx.l,ctx.n,g.device)
        _gx=_SO3partB_array.view(gx)
        gelib_base.add_conterpolate3dB_back(_gx,_g,_M)
        return gx,None



## ----------------------------------------------------------------------------------------------------------
## ---- Helpers ---------------------------------------------------------------------------------------------
## ----------------------------------------------------------------------------------------------------------


def CGproduct(x,y,maxl=-1):
    return x.CGproduct(y,maxl)
    





# This file is part of GElib, a C++/CUDA library for group equivariant 
# tensor operations. 
#  
# Copyright (c) 2023, Imre Risi Kondor
#
# This source code file is subject to the terms of the noncommercial 
# license distributed with GElib in the file NONCOMMERICAL.TXT. Commercial 
# use is prohibited. All redistributed versions of this file (in orginal
# or modified form) must retain this copyright notice and must be 
# accompanied by a verbatim copy of the license. 

import torch
from . import gelib_base as gb


class SO3element(torch.Tensor):
    """
    An element of the group SO(3)
    """

    def __init__(self, _T):
        self=_T
        #super().__init__()


    # ---- Static constructors -----------------------------------------------------------------------------


    @classmethod
    def identity(self):
        return SO3element(torch.eye(3))

    @classmethod
    def random(self):
        return SO3element(gb.SO3element.random().torch())


    # ---- Operations --------------------------------------------------------------------------------------


    def __mul__(self,y):
        assert(isinstance(y,SO3element))
        return SO3element(torch.matmul(self,y))

    def inv(self):
        return SO3element(self.t())


    # ---- I/O ----------------------------------------------------------------------------------------------



# This file is part of GElib, a C++/CUDA library for group equivariant 
# tensor operations. 
#  
# Copyright (c) 2025, Imre Risi Kondor
#
# This source code file is subject to the terms of the noncommercial 
# license distributed with GElib in the file NONCOMMERICAL.TXT. Commercial 
# use is prohibited. All redistributed versions of this file (in orginal
# or modified form) must retain this copyright notice and must be 
# accompanied by a verbatim copy of the license. 

import torch
from . import gelib_base as gb

class gather_map:

    def __init__(self, x):
        assert isinstance(x,gb.gather_map)
        self.obj=x

    @classmethod
    def from_matrix(self,M,n_in,n_out):
        assert isinstance(M,torch.Tensor)
        assert isinstance(n_in,int)
        assert isinstance(n_out,int)
        return gather_map(gb.gather_map.from_matrix(x.int(),n_in,n_out))

    @classmethod
    def random(self,n_in,n_out,p):
        return gather_map(gb.gather_map.random(n_in,n_out,p))



    # ---- I/O ----------------------------------------------------------------------------------------------


    def __repr__(self):
        return self.obj.__repr__()

    def __str__(self):
        return self.obj.__str__()




import os
import torch
from setuptools import setup
from setuptools import find_packages
from torch.utils.cpp_extension import CppExtension, BuildExtension, CUDAExtension
import time
from os.path import splitext
from os.path import basename
from glob import glob
from typing import Union, Tuple


def interpret_bool_string(string:Union[str,bool], _true_values:Tuple[str] = ("TRUE", "ON"), _false_values:Tuple[str] = ("FALSE", "OFF")):
    if isinstance(string, bool):
        return string
    if string.strip().upper() in _true_values:
        return True
    if string.strip().upper() in _false_values:
        return False
    raise ValueError(f"String {string} cannot be interpreted as True or False. Any upper/lower-case version of {_true_values} is True, {_false_values} is False. {string} was neither.")



def main():

    compile_with_cuda = interpret_bool_string(os.environ.get("WITH_CUDA", False))

    copy_warnings = False
    torch_convert_warnings = True

    cwd = os.getcwd()
    cnine_folder = "/deps/cnine"
    ext_cuda_folder = "cuda/"

    _include_dirs = [cwd + cnine_folder + '/include',
		     cwd + cnine_folder + '/combinatorial',
		     cwd + cnine_folder + '/containers',
		     cwd + cnine_folder + '/hpc',
		     cwd + cnine_folder + '/math',
		     cwd + cnine_folder + '/matrices',
		     cwd + cnine_folder + '/utility',
		     cwd + cnine_folder + '/tensors',
		     cwd + cnine_folder + '/tensors/functions',
		     cwd + cnine_folder + '/wrappers',
                     cwd + cnine_folder + '/include/cmaps',
                     cwd + cnine_folder + '/objects/matrix',
                     cwd + cnine_folder + '/tensor_views',
                     cwd + '/include',
                     cwd + '/cuda',
                     cwd + '/core',
                     cwd + '/SO3',
                     cwd + '/O3'
                     ]


    _cxx_compile_args = ['-std=c++17',
                         '-Wno-sign-compare',
                         '-Wno-deprecated-declarations',
                         '-Wno-unused-variable',
                         '-Wno-reorder',
                         '-Wno-reorder-ctor',
                         '-Wno-overloaded-virtual',
                         '-D_WITH_ATEN',
                         '-DCNINE_RANGE_CHECKING',
                         '-DCNINE_SIZE_CHECKING',
                         '-DCNINE_DEVICE_CHECKING',
                         '-DGELIB_RANGE_CHECKING',
                         '-DWITH_FAKE_GRAD'
                         ]

    _nvcc_compile_args = ['-D_WITH_CUDA',
                          '-D_WITH_CUBLAS',
                          '-D_DEF_CGCMEM',
                          '-DGELIB_RANGE_CHECKING',
                          '-DWITH_FAKE_GRAD',
                          '-std=c++17',
                          '-rdc=true'
                          ]

    if copy_warnings:
        _cxx_compile_args.extend([
            '-DCNINE_COPY_WARNINGS',
            '-DCNINE_ASSIGN_WARNINGS',
            '-DCNINE_MOVE_WARNINGS',
            '-DCNINE_MOVEASSIGN_WARNINGS',
            '-DGELIB_COPY_WARNINGS',
            '-DGELIB_MOVE_WARNINGS',
            '-DGELIB_CONVERT_WARNINGS'
        ])

    if torch_convert_warnings:
        _cxx_compile_args.extend([
            '-DCNINE_ATEN_CONVERT_WARNINGS'
        ])

    if compile_with_cuda:
        _cxx_compile_args.extend(['-D_WITH_CUDA', '-D_WITH_CUBLAS'])

    _depends = ['setup.py',
                'src/gelib.cpp',
                'bindings/*.cpp'
                ]


    # ---- Compilation commands ----------------------------------------------------------------------------------


    if compile_with_cuda:
        ext_modules = [CUDAExtension('gelib_base', [
            cnine_folder+'include/Cnine_base.cu',
            cnine_folder+'cuda/TensorView_assign.cu',
            'cuda/GElib_base.cu',
            'cuda/SO3part_addCGproduct.cu',
            'cuda/SO3part_addCGproduct_back0.cu',
            'cuda/SO3part_addCGproduct_back1.cu',
            'python/bindings/GElib_py.cpp'
        ],
            include_dirs=_include_dirs,
            extra_compile_args={
            'nvcc': _nvcc_compile_args,
            'cxx': _cxx_compile_args},
            depends=_depends
        )]
    else:
        ext_modules = [CppExtension('gelib_base', ['python/bindings/GElib_py.cpp'],
                                    include_dirs=_include_dirs,
                                    extra_compile_args={
            'cxx': _cxx_compile_args},
            depends=_depends
        )]

    setup(name='gelib',
          ext_modules=ext_modules,
          packages=find_packages('python/src'),
          package_dir={'': 'python/src'},
          py_modules=[splitext(basename(path))[0] for path in glob('python/src/*.py')],
          include_package_data=True,
          zip_safe=False,
          cmdclass={'build_ext': BuildExtension})

    print("Compilation finished:", time.ctime(time.time()))

    # ------------------------------------------------------------------------------------------------------------


if __name__ == "__main__":
    main()



# This file is part of GElib, a C++/CUDA library for group
# equivariant tensor operations. 
# 
# Copyright (c) 2022, Imre Risi Kondor
#
# This Source Code Form is subject to the terms of the Mozilla
# Public License v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import torch
from . import gelib_base as gb
from gelib import *


def common_batch(x,y):
    assert x.dim()>=1
    assert y.dim()>=1
    if x.size(0)>1 and y.size(0)>1:
        assert(x.size(0)==y.size(0))
    return max(x.size(0),y.size(0))

